{
/**
 * This library provides formatted string output methods based on the
 * standard C sprintf function.  With this class you can convert
 * byte, word and long types into binary, octal, decimal or hexadecimal formatted
 * strings.  You can specify the minimum and maximum number of columns to
 * display your values, and these values can be left or right justified, with
 * or without padded zeros.  Methods are included for the following C library
 * functions: itoa, atoi and sprintf.
 *
 * This implementation defines the function bprintf that take one,
 * and only one, variable parameter. The original format string must be split
 * into pieces that all have one format specifier. This normally is not a
 * problem since most format strings consists of fixed text with format
 * specifiers for values.
 *
 * @version 1.2 Dec 16, 2007
 * @author Peter Verkaik (verkaik6@zonnet.nl)
 */
}


DAT
  buf byte 0[33] '// for integer conversions (base2 is 32 binary digits plus closing null)

  
PUB isdigit(ch): YesNo
  '/**
  ' * Test if character is a digit.
  ' *
  ' * @param ch Character to be tested
  ' * @return True if ch is a digit
  ' */
  YesNo := (ch =< "9") AND (ch => "0")


PUB isspace(ch): YesNo
  '/**
  ' * Test if character is a space (0x20), tab (\t) or newline (\n).
  ' *
  ' * @param ch Character to be tested
  ' * @return True if ch is a space, tab or newline
  ' */
  YesNo := (ch == 32) OR (ch == 9) OR (ch == 10)


PUB reverse(s) | c,k
  '/**
  ' * Reverse string in place.
  ' *
  ' * @param s Character array to reverse
  ' */
  k := s + strsize(s) - 1 'address of last character
  repeat while (s < k)
    c := byte[s]
    byte[s++] := byte[k]
    byte[k--] := c


PUB itoa(n,s) | sign,sorg
  '/**
  ' * Convert signed integer to signed decimal string.
  ' *
  ' * @param n Integer value to convert
  ' * @param s Character array to hold output
  ' */
  sorg := s
  sign := n                              '// record sign
  if n < 0                              
    n := -n                              '// make n positive
  repeat                                 '// generate digits in reverse order
    byte[s++] := ((n//10) + "0")         '// get next digit
    n := n/10
  until n == 0
  if sign < 0
    byte[s++] := "-"                     '// sign character only if negative
  byte[s] := 0                           '// closing 0
  reverse(sorg)


PUB atoi(s): val | sign,n
  '/**
  ' * Convert signed decimal string to signed integer.
  ' *
  ' * @param s Character array that holds decimal string
  ' * @return Value of signed decimal string
  ' */
  repeat while isspace(byte[s])
    ++s                                  '// skip over any white space
  sign := 1                              '// assume positive
  case byte[s]
    "-": sign := -1                      '// is negative
         ++s                             '// advance only if sign (- or +) present
    "+": ++s
  n := 0
  repeat while isdigit(byte[s])
    n := (10*n) + (byte[s++] - "0")      '// calculate value
  val := sign * n                        '// adjust for sign

 
PUB sprintf(str,fmt,arg): index
  '/**
  ' * Print formatted string to character buffer.
  ' * This sprintf only accepts one argument to print.
  ' * See bprintf how to print multiple arguments.
  ' *
  ' * format specifiers for byte, word, long: %c,%d,%i,%b,%o,%u,%x
  ' * format specifier for string: %s
  ' * field specification (example)
  ' *   %-05.7d where - for left justify, default right justify
  ' *                 0 for padding with zeroes, default spaces
  ' *                 5 minimum field width
  ' *                 . field separator
  ' *                 7 maximum field width
  ' *
  ' * An original C sprintf statement like:
  ' * sprintf(buffer,"outside temperature %d celsius",24);
  ' * would be written as:
  ' * OBJ
  ' *   fmt: "Format"
  ' * VAR
  ' *   byte buffer[128]
  ' * PUB
  ' *   fmt.sprintf(@buffer,string("outside temperature %d celsius"),24)
  ' *
  ' * @param str Character array for formatted output
  ' * @param fmt String defining format
  ' * @param arg value parameter
  ' * @return Index in str pointing beyond last written character (eg. size of output string)
  ' */
  index := bprintf(str,0,fmt,arg)
  byte[str+index] := 0


PUB bprintf(str,si,fmt,arg): index | pad,len,maxlen,minlen,bi,left,strtype,sorg
  '/**
  ' * Print formatted string to character buffer, maintaining buffer index.
  ' * (special version of sprintf, to ease porting printf and sprintf with multiple arguments)
  ' *
  ' * format specifiers for byte, word, long: %c,%d,%i,%b,%o,%u,%x
  ' * format specifier for string: %s
  ' * field specification (example)
  ' *   %-05.7d where - for left justify, default right justify
  ' *                 0 for padding with zeroes, default spaces
  ' *                 5 minimum field width
  ' *                 . field separator
  ' *                 7 maximum field width
  ' *
  ' * An original C sprintf statement like:
  ' * sprintf(buffer,"outside temperature %d %s inside temperature %d %s",12,"fahrenheit",24,"celsius");
  ' * would be written as:
  ' * OBJ
  ' *   fmt: "Format"
  ' * VAR
  ' *   byte buffer[128]
  ' *   word k
  ' * PUB
  ' *   k := fmt.bprintf(@buffer,0,string("outside temperature %d "),12)
  ' *   k := fmt.bprintf(@buffer,k,string("%s "),string("fahrenheit"))
  ' *   k := fmt.bprintf(@buffer,k,string("inside temperature %d "),24)
  ' *   k := fmt.bprintf(@buffer,k,string("%s"),string("celsius"))
  ' *   buffer[k] := 0
  ' * The original sprintf is simply split up in smaller bprintf statements that take only 1 argument.
  ' * The end result is the same: a single output string in buffer.
  ' *
  ' * @param str Character array for formatted output
  ' * @param si Start index in str
  ' * @param fmt String defining format
  ' * @param arg value parameter
  ' * @return Index in str pointing beyond last written character
  ' */
  sorg := str 'save start address of output buffer
  len := 0
  str += si 'start from index
  repeat while byte[fmt] <> 0
    if byte[fmt] <> "%"
      if byte[fmt] <> "\"
        byte[str++] := byte[fmt++]
      else
        fmt++ 'skip backslash
        case byte[fmt]
          "\": byte[str++] := "\"
               fmt++
          "t": byte[str++] := $09
               fmt++
          "n": byte[str++] := $0A
               fmt++
          "r": byte[str++] := $0D
               fmt++
          other: byte[str++] := "\" 'output skipped backslash
                 byte[str++] := byte[fmt++]
      next
    else
      fmt++ 'skip control character %
    if byte[fmt] == "%"
      byte[str++] := byte[fmt++] '%%
      next
    if byte[fmt] == "-"
      left := true
      ++fmt 'skip -
    else
      left := false
    if byte[fmt] == "0"
      pad := "0"
    else
      pad := " "
    if isdigit(byte[fmt])  '// minimum field width
      bi := 0
      repeat while isdigit(byte[fmt])
        buf[bi++] := byte[fmt++]
      buf[bi] := 0
      minlen := atoi(@buf)
    else
      minlen := 0
    if byte[fmt] == "."   '// maximum field width
      ++fmt 'skip .
      bi := 0
      repeat while isdigit(byte[fmt])
        buf[bi++] := byte[fmt++]
      buf[bi] := 0
      maxlen := atoi(@buf)
    else
      maxlen := 0
    strtype := false 'assume no string value
    case byte[fmt++]
      "c": buf[0] := arg       '// character
           buf[1] := 0
           len := 1
      "d","i": itoa(arg, @buf)  '// signed decimal
               len := strsize(@buf)
      "b": itoab(arg, @buf, 2)  '// binary
           len := strsize(@buf)
      "o": itoab(arg, @buf, 8)  '// octal
           len := strsize(@buf)
      "u": itoab(arg, @buf, 10) '// unsigned decimal
           len := strsize(@buf)
      "x": itoab(arg, @buf, 16) '// hexadecimal
           len := strsize(@buf)
      "s": strtype := true
           len := strsize(arg) '// string
      other: buf[0] := 0       '// no valid format specifier
             len := 0
    if (maxlen <> 0) AND (maxlen < len)
      len := maxlen
    if (minlen > len)
      minlen := minlen - len
    else
      minlen := 0
    bi := 0
    if left == false
      if (buf[bi] == "-") AND (pad == "0")
        byte[str++] := buf[bi++]
        len--
      repeat while minlen > 0
        minlen--
        byte[str++] := pad
    repeat while len > 0
      len--
      if strtype == false
        byte[str++] := buf[bi++]  'copy ascii string of value
      else
        if (byte[arg] == 0)
          quit
        byte[str++] := byte[arg++]  'copy string argument
    if left == true
      repeat while minlen > 0
        minlen--
        byte[str++] := pad
  return str-sorg


PUB itoab(n,s,b) | lowbit,sorg
  '/*
  ' * Convert unsigned integer to unsigned string using specific base.
  ' * (This is a non-standard function). Used by <code>printf</code>.
  ' *
  ' * @param n Integer value to convert
  ' * @param s Character array to hold output
  ' * @param b Base for output (2=binary, 8=octal, 10=decimal, 16=hexadecimal)
  ' */
  sorg := s
  b := (b >> 1) & $F
  repeat                             '// generate digits/letters in reverse order
    lowbit := n & 1
    n := (n >> 1) & $7FFFFFFF
    byte[s] := ((n // b) << 1) + lowbit
    if (byte[s] < 10)
      byte[s] += "0"
    else
      byte[s] += ("A"-10)
    ++s
    n := n/b
  until (n == 0)
  byte[s] := 0                        '// closing 0
  reverse(sorg)

  